package types

import (
	"database/sql"
	"database/sql/driver"
	"electricity_bill_calc/tools"
	"encoding/json"
	"errors"

	"github.com/jackc/pgx/v5/pgtype"
)

type DateTimeRange struct {
	pgtype.Range[DateTime]
}

func NewEmptyDateTimeRange() DateTimeRange {
	return DateTimeRange{
		Range: pgtype.Range[DateTime]{
			Lower:     MinDateTime(),
			LowerType: pgtype.Unbounded,
			Upper:     MaxDateTime(),
			UpperType: pgtype.Unbounded,
			Valid:     true,
		},
	}
}

func NewDateTimeRange(lower *DateTime, upper *DateTime) DateTimeRange {
	return DateTimeRange{
		Range: pgtype.Range[DateTime]{
			LowerType: tools.Cond(lower != nil, pgtype.Inclusive, pgtype.Unbounded),
			Lower:     tools.DefaultTo(lower, MinDateTime()),
			UpperType: tools.Cond(upper != nil, pgtype.Inclusive, pgtype.Unbounded),
			Upper:     tools.DefaultTo(upper, MaxDateTime()),
			Valid:     true,
		},
	}
}

var _ driver.Value = (*DateTimeRange)(nil)

func (dr DateTimeRange) Value() (driver.Value, error) {
	return assembleRange(dr.Range), nil
}

var _ sql.Scanner = (*DateTimeRange)(nil)

func (dr *DateTimeRange) Scan(src interface{}) (err error) {
	switch src := src.(type) {
	case pgtype.Range[DateTime]:
		dr.Range = src
	case string:
		r, err := destructureToRange[DateTime](src)
		if err != nil {
			return err
		}
		dr.Range = r
	case []byte:
		r, err := destructureToRange[DateTime](string(src))
		if err != nil {
			return err
		}
		dr.Range = r
	case nil:
		dr = nil
	default:
		return errors.New("该数据类型不支持解析到日期范围。")
	}
	return
}

var _ json.Marshaler = (*DateTimeRange)(nil)

func (dr DateTimeRange) MarshalJSON() ([]byte, error) {
	return json.Marshal(assembleRange(dr.Range))
}

var _ json.Unmarshaler = (*DateTimeRange)(nil)

func (dr *DateTimeRange) UnmarshalJSON(data []byte) error {
	var str string
	if err := json.Unmarshal(data, &str); err != nil {
		return errors.New("不能解析指定的日期范围值。")
	}
	r, err := destructureToRange[DateTime](str)
	if err != nil {
		return err
	}
	dr.Range = r
	return nil
}

func (dr *DateTimeRange) SetLower(lower DateTime, bound ...pgtype.BoundType) {
	bound = append(bound, pgtype.Inclusive)
	dr.Range.Lower = lower
	dr.Range.LowerType = bound[0]
}

func (dr *DateTimeRange) SetLowerUnbounded() {
	dr.Range.Lower = MinDateTime()
	dr.Range.LowerType = pgtype.Unbounded
}

func (dr *DateTimeRange) SetUpper(upper DateTime, bound ...pgtype.BoundType) {
	bound = append(bound, pgtype.Inclusive)
	dr.Range.Upper = upper
	dr.Range.UpperType = bound[0]
}

func (dr *DateTimeRange) SetUpperUnbounded() {
	dr.Range.Upper = MaxDateTime()
	dr.Range.UpperType = pgtype.Unbounded
}

func (dr DateTimeRange) IsEmptyOrWild() bool {
	return (dr.Range.LowerType == pgtype.Unbounded && dr.Range.UpperType == pgtype.Unbounded) ||
		(dr.Range.LowerType == pgtype.Empty || dr.Range.UpperType == pgtype.Empty)
}
